<?php

namespace Intuit\Response;

use Illuminate\Support\Arr;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\App;
use Illuminate\Support\Facades\Log;

use Intuit\Models\{Category,Customer,Invoice,Item};

/**
 * This is a Generic Intuit Response to API calls that produces a list of objects
 */
class ListList extends Base implements \Countable, \ArrayAccess, \Iterator
{
	protected const LOGKEY = 'RLI';

	protected Collection $_data;
	protected ?string $_type;
	protected int $startPosition;
	protected int $maxResults;

	private ?int $counter = NULL;
	protected const TYPES = [
		'classes' => 'Class',
		'customers' => 'Customer',
		'invoices' => 'Invoice',
		'items' => 'Item',
	];

	/**
	 * Default Constructor Setup
	 *
	 * @param object $response
	 * @param string $type
	 * @throws \Exception
	 */
	public function __construct(object $response,string $type)
	{
		parent::__construct($response);

		if (! in_array($type,array_keys(self::TYPES)))
			throw new \Exception('Unknown data type: '.$type);

		if (object_get($response,'time'))
			unset($response->time);

		$this->_data = $this->data($response,$type);

		// This is only for child classes
		if (get_class($this) == Base::class) {
			Log::debug(sprintf('%s:Intuit RESPONSE Initialised [%s]',static::LOGKEY,get_class($this)),['m'=>__METHOD__]);

			if (App::environment() == 'dev')
				file_put_contents('/tmp/response',print_r($this,TRUE),FILE_APPEND);
		}
	}

	public function __call(string $name,array $args)
	{
		switch ($name) {
			case 'first':
				return $this->_data->{$name}();
			case 'pluck':
				return $this->_data->map(function($item) use ($args) { return $item->{$args[0]}; });

			default:
				throw new \BadMethodCallException(sprintf('Call to undefined method %s::%s()',get_class($this),$name));
		}
	}

	public function current(): mixed
	{
		return $this->_data[$this->counter];
	}

	public function next(): void
	{
		$this->counter++;
	}

	public function key(): mixed
	{
		return $this->counter;
	}

	public function valid(): bool
	{
		return isset($this->_data[$this->counter]);
	}

	public function rewind(): void
	{
		$this->counter = 0;
	}

	public function offsetExists(mixed $offset): bool
	{
		return $this->_data->has($offset);
	}

	public function offsetGet(mixed $offset): mixed
	{
		return $this->_data->get($offset);
	}

	public function offsetSet(mixed $offset, mixed $value): void
	{
		throw new \Exception('Method not implemented: '.__METHOD__);
	}

	public function offsetUnset(mixed $offset): void
	{
		$this->_data->forget($offset);

		// Rekey the collection
		$this->_data = $this->_data->values();

		// Reset the counter if we have deleted a value before it
		if ($offset < $this->counter)
			$this->counter--;
	}

	public function count(): int
	{
		return $this->_data->count();
	}

	/* METHODS */

	/**
	 * Convert our response into a collection of the appropriate model
	 *
	 * @param object $response
	 * @param string $type
	 * @return Collection
	 * @throws \Exception
	 */
	private function data(object $response,string $type): Collection
	{
		if (! ($x=object_get($response->QueryResponse,Arr::get(self::TYPES,$type))))
			return collect();

		switch ($type) {
			case 'classes':
				$data = collect(Category::hydrate($x));
				break;

			case 'customers':
				$data = collect(Customer::hydrate($x));
				break;

			case 'invoices':
				$data = collect(Invoice::hydrate($x));
				break;

			case 'items':
				$data = collect(Item::hydrate($x));
				break;

			default: throw new \Exception('Unknown object type: '.$this->_type);
		}

		$this->startPosition = $response->QueryResponse->startPosition;
		$this->maxResults = $response->QueryResponse->maxResults;

		return $data;
	}
}